# ! Basic necessities to get the CLI running
import signal
import sys
import argparse
import pkg_resources
import os

# ! The actual download stuff
from spotdl.download.downloader import DownloadManager
from spotdl.search.spotifyClient import SpotifyClient
import spotdl.search.songGatherer as songGatherer

from spotdl.download import ffmpeg

# ! Usage is simple - call:
#   'python __main__.py <links, search terms, tracking files separated by spaces>
# ! Eg.
# !      python __main__.py
# !          https://open.spotify.com/playlist/37i9dQZF1DWXhcuQw7KIeM
# !          'old gods of asgard Control'
# !          https://open.spotify.com/album/2YMWspDGtbDgYULXvVQFM6
# !          https://open.spotify.com/track/08mG3Y1vljYA6bvDt4Wqkj
# !
# ! Well, yeah its a pretty long example but, in theory, it should work like a charm.
# !
# ! A '.spotdlTrackingFile' is automatically  created with the name of the first song in the
# ! playlist/album or the name of the song supplied. We don't really re re re-query YTM and Spotify
# ! as all relevant details are stored to disk.
# !
# ! Files are cleaned up on download failure.
# !
# ! All songs are normalized to standard base volume. the soft ones are made louder,
# ! the loud ones, softer.
# !
# ! The progress bar is synched across multiple-processes (4 processes as of now), getting the
# ! progress bar to synch was an absolute pain, each process knows how much 'it' progressed,
# ! but the display has to be for the overall progress so, yeah... that took time.
# !
# ! spotdl will show you its true speed on longer download's - so make sure you try
# ! downloading a playlist.
# !
# ! still yet to try and package this but, in theory, there should be no errors.
# !
# !                                                          - cheerio! (Michael)
# !
# ! P.S. Tell me what you think. Up to your expectations?

# ! Script Help
help_notice = """
To download a song run,
    spotdl [trackUrl]
    ex. spotdl https://open.spotify.com/track/0VjIjW4GlUZAMYd2vXMi3b

To download a album run,
    spotdl [albumUrl]
    ex. spotdl https://open.spotify.com/album/4yP0hdKOZPNshxUOjY0cZj

To download a playlist, run:
    spotdl [playlistUrl]
    ex. spotdl https://open.spotify.com/playlist/37i9dQZF1E8UXBoz02kGID

To download your saved songs, run:
    spotdl --user-auth saved

To download all songs from an artist run:
    spotdl [artistUrl]
    ex. spotdl https://open.spotify.com/artist/1fZAAHNWdSM5gqbi9o5iEA

To change output format run:
    spotdl [songUrl] --output-format mp3/m4a/flac/opus/ogg
    ex. spotdl [songUrl] --output-format opus

To use ffmpeg binary that is not on PATH run:
    spotdl [songUrl] --ffmpeg path/to/your/ffmpeg.exe
    ex. spotdl [songUrl] --ffmpeg C:\ffmpeg\bin\ffmpeg.exe

To ignore your ffmpeg version run:
    spotdl --ignore-ffmpeg-version

To search for and download a song, run, with quotation marks:
Note: This is not accurate and often causes errors.
    spotdl [songQuery]
    ex. spotdl 'The Weeknd - Blinding Lights'

To resume a failed/incomplete download, run:
    spotdl [pathToTrackingFile]
    ex. spotdl 'The Weeknd - Blinding Lights.spotdlTrackingFile'

    .spotdlTrackingFiles are automatically created when a download starts and deleted on completion

You can queue up multiple download tasks by separating the arguments with spaces:
    spotdl [songQuery1] [albumUrl] [songQuery2] ... (order does not matter)
    ex. spotdl 'The Weeknd - Blinding Lights'
            https://open.spotify.com/playlist/37i9dQZF1E8UXBoz02kGID ...

You can use the --debug-termination flag to figure out where in the code spotdl got stuck.

spotDL downloads up to 4 songs in parallel, so for a faster experience,
download albums and playlist, rather than tracks.
"""


def console_entry_point():
    """
    This is where all the console processing magic happens.
    Its super simple, rudimentary even but, it's dead simple & it works.
    """

    if len(sys.argv) >= 2 and sys.argv[1] in ["-v", "--version"]:
        version = pkg_resources.require("spotdl")[0].version
        print(version)
        sys.exit(0)

    arguments = parse_arguments()
    args_dict = vars(arguments)

    if (
        ffmpeg.has_correct_version(
            arguments.ignore_ffmpeg_version, arguments.ffmpeg or "ffmpeg"
        )
        is False
    ):
        sys.exit(1)

    for request in arguments.query:
        if "saved" == request and not arguments.userAuth:
            arguments.userAuth = True
            print(
                "Detected 'saved' in command line, but no --user-auth flag. Enabling Anyways."
            )
            print("Please Log In...")

    SpotifyClient.init(
        client_id="5f573c9620494bae87890c0f08a60293",
        client_secret="212476d9b0f3472eaa762d90b19b0ba8",
        user_auth=arguments.userAuth,
    )

    if arguments.path:
        if not os.path.isdir(arguments.path):
            sys.exit("The output directory doesn't exist.")
        print(f"Will download to: {os.path.abspath(arguments.path)}")
        os.chdir(arguments.path)

    with DownloadManager(args_dict) as downloader:
        if not arguments.debug_termination:

            def gracefulExit(signal, frame):
                downloader.displayManager.close()
                sys.exit(0)

            signal.signal(signal.SIGINT, gracefulExit)
            signal.signal(signal.SIGTERM, gracefulExit)

        songObjList = []
        for request in arguments.query:
            if request.endswith(".spotdlTrackingFile"):
                print("Preparing to resume download...")
                downloader.resume_download_from_tracking_file(request)
            else:
                songObjList.extend(songGatherer.from_query(request, arguments.format))
                # linefeed to visually separate output for each query
                print()

        if len(songObjList) > 0:
            downloader.download_multiple_songs(songObjList)


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog="spotdl",
        description=help_notice,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "query", type=str, nargs="+", help="URL/String for a song/album/playlist/artist"
    )
    parser.add_argument("--debug-termination", action="store_true")
    parser.add_argument("-o", "--output", help="Output directory path", dest="path")
    parser.add_argument(
        "-of",
        "--output-format",
        help="Output format",
        dest="format",
        choices={"mp3", "m4a", "flac", "ogg", "opus"},
        default="mp3",
    )
    parser.add_argument(
        "--user-auth",
        help="Use User Authentication",
        action="store_true",
        dest="userAuth",
    )
    parser.add_argument("-f", "--ffmpeg", help="Path to ffmpeg", dest="ffmpeg")
    parser.add_argument(
        "--ignore-ffmpeg-version", help="Ignore ffmpeg version", action="store_true"
    )
    parser.add_argument(
        "-v",
        "--version",
        help="Show spotDL's version and exit",
        action="store_true",
    )

    return parser.parse_args()


if __name__ == "__main__":
    console_entry_point()



# ===============
# === Imports ===
# ===============
import asyncio
import concurrent.futures
from spotdl.download.embed_metadata import set_id3_data
import sys
import traceback

from pathlib import Path

# ! The following are not used, they are just here for static typechecking with mypy
from typing import List, Optional

from pytube import YouTube

from spotdl.download.progressuiHandlers import DisplayManager
from spotdl.download.trackingfileHandlers import DownloadTracker
from spotdl.search.songObj import SongObj
from spotdl.download import ffmpeg


# ==========================
# === Base functionality ===
# ==========================

# ========================
# === Helper function ===
# ========================

def _sanitize_filename(input_str: str) -> str:
    output = input_str

    # ! this is windows specific (disallowed chars)
    output = "".join(char for char in output if char not in "/?\\*|<>")

    # ! double quotes (") and semi-colons (:) are also disallowed characters but we would
    # ! like to retain their equivalents, so they aren't removed in the prior loop
    output = output.replace('"', "'").replace(':', '-')

    return output


def _get_smaller_file_path(input_song: SongObj, output_format: str) -> Path:
    # Only use the first artist if the song path turns out to be too long
    smaller_name = f"{input_song.get_contributing_artists()[0]} - {input_song.get_song_name()}"

    # ! this is windows specific (disallowed chars)
    smaller_name = "".join(char for char in smaller_name if char not in "/?\\*|<>")

    # ! double quotes (") and semi-colons (:) are also disallowed characters
    # ! but we would like to retain their equivalents, so they aren't removed
    # ! in the prior loop
    smaller_name = smaller_name.replace('"', "'")
    smaller_name = smaller_name.replace(':', '-')

    smaller_name = _sanitize_filename(smaller_name)

    try:
        return Path(f"{smaller_name}.{output_format}").resolve()
    except (OSError, WindowsError):
        # Expected to happen in the rare case when the saved path is too long,
        # even with the short filename
        raise OSError("Cannot save song due to path issues.")


def _get_converted_file_path(song_obj: SongObj, output_format: str = None) -> Path:

    # ! we eliminate contributing artist names that are also in the song name, else we
    # ! would end up with things like 'Jetta, Mastubs - I'd love to change the world
    # ! (Mastubs REMIX).mp3' which is kinda an odd file name.

    # also make sure that main artist is included in artistStr even if they
    # are in the song name, for example
    # Lil Baby - Never Recover (Lil Baby & Gunna, Drake).mp3

    artists_filtered = []

    if output_format is None:
        output_format = "mp3"

    for artist in song_obj.get_contributing_artists():
        if artist.lower() not in song_obj.get_song_name():
            artists_filtered.append(artist)
        elif artist.lower() is song_obj.get_contributing_artists()[0].lower():
            artists_filtered.append(artist)

    artist_str = ", ".join(artists_filtered)

    converted_file_name = _sanitize_filename(
        f"{artist_str} - {song_obj.get_song_name()}.{output_format}"
    )

    converted_file_path = Path(converted_file_name)

    # ! Checks if a file name is too long (256 max on both linux and windows)
    try:
        if len(str(converted_file_path.resolve().name)) > 256:
            print("Path was too long. Using Small Path.")
            return _get_smaller_file_path(song_obj, output_format)
    except (OSError, WindowsError):
        return _get_smaller_file_path(song_obj, output_format)

    return converted_file_path


# ===========================================================
# === The Download Manager (the tyrannical boss lady/guy) ===
# ===========================================================


class DownloadManager:
    # ! Big pool sizes on slow connections will lead to more incomplete downloads
    poolSize = 4

    def __init__(self, arguments: Optional[dict] = None):
        # start a server for objects shared across processes
        self.displayManager = DisplayManager()
        self.downloadTracker = DownloadTracker()

        if sys.platform == "win32":
            # ! ProactorEventLoop is required on Windows to run subprocess asynchronously
            # ! it is default since Python 3.8 but has to be changed for previous versions
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)

        self.loop = asyncio.get_event_loop()
        # ! semaphore is required to limit concurrent asyncio executions
        self.semaphore = asyncio.Semaphore(self.poolSize)

        # ! thread pool executor is used to run blocking (CPU-bound) code from a thread
        self.thread_executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.poolSize
        )

        if arguments is None:
            arguments = {}

        arguments.setdefault("ffmpeg_path", "ffmpeg")
        arguments.setdefault("format", "mp3")

        self.arguments = arguments

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.displayManager.close()

    def download_single_song(self, songObj: SongObj) -> None:
        """
        `songObj` `song` : song to be downloaded

        RETURNS `~`

        downloads the given song
        """

        self.downloadTracker.clear()
        self.downloadTracker.load_song_list([songObj])

        self.displayManager.set_song_count_to(1)

        self._download_asynchronously([songObj])

    def download_multiple_songs(self, songObjList: List[SongObj]) -> None:
        """
        `list<songObj>` `songObjList` : list of songs to be downloaded

        RETURNS `~`

        downloads the given songs in parallel
        """

        self.downloadTracker.clear()
        self.downloadTracker.load_song_list(songObjList)

        self.displayManager.set_song_count_to(len(songObjList))

        self._download_asynchronously(songObjList)

    def resume_download_from_tracking_file(self, trackingFilePath: str) -> None:
        """
        `str` `trackingFilePath` : path to a .spotdlTrackingFile

        RETURNS `~`

        downloads songs present on the .spotdlTrackingFile in parallel
        """

        self.downloadTracker.clear()
        self.downloadTracker.load_tracking_file(trackingFilePath)

        songObjList = self.downloadTracker.get_song_list()

        self.displayManager.set_song_count_to(len(songObjList))

        self._download_asynchronously(songObjList)

    def _download_asynchronously(self, song_obj_list):
        tasks = [self._pool_download(song) for song in song_obj_list]
        # call all task asynchronously, and wait until all are finished
        self.loop.run_until_complete(asyncio.gather(*tasks))

    async def _pool_download(self, song_obj: SongObj):
        # ! Run asynchronous task in a pool to make sure that all processes
        # ! don't run at once.

        # tasks that cannot acquire semaphore will wait here until it's free
        # only certain amount of tasks can acquire the semaphore at the same time
        async with self.semaphore:
            return await self.download_song(song_obj)

    async def download_song(self, songObj: SongObj) -> None:
        """
        `songObj` `songObj` : song to be downloaded

        RETURNS `~`

        Downloads, Converts, Normalizes song & embeds metadata as ID3 tags.
        """

        dispayProgressTracker = self.displayManager.new_progress_tracker(songObj)

        # ! since most errors are expected to happen within this function, we wrap in
        # ! exception catcher to prevent blocking on multiple downloads
        try:

            # ! all YouTube downloads are to .\Temp; they are then converted and put into .\ and
            # ! finally followed up with ID3 metadata tags

            # ! we explicitly use the os.path.join function here to ensure download is
            # ! platform agnostic

            # Create a .\Temp folder if not present
            tempFolder = Path(".", "Temp")

            if not tempFolder.exists():
                tempFolder.mkdir()

            convertedFilePath = _get_converted_file_path(songObj, self.arguments["format"])

            # if a song is already downloaded skip it
            if convertedFilePath.is_file():
                if self.displayManager:
                    dispayProgressTracker.notify_download_skip()
                if self.downloadTracker:
                    self.downloadTracker.notify_download_completion(songObj)

                # ! None is the default return value of all functions, we just explicitly define
                # ! it here as a continent way to avoid executing the rest of the function.
                return None

            # download Audio from YouTube
            if dispayProgressTracker:
                youtubeHandler = YouTube(
                    url=songObj.get_youtube_link(),
                    on_progress_callback=dispayProgressTracker.pytube_progress_hook,
                )

            else:
                youtubeHandler = YouTube(songObj.get_youtube_link())

            trackAudioStream = (
                youtubeHandler.streams.filter(only_audio=True)
                .order_by("bitrate")
                .last()
            )
            if not trackAudioStream:
                print(
                    f'Unable to get audio stream for "{songObj.get_song_name()}" '
                    f'by "{songObj.get_contributing_artists()[0]}" '
                    f'from video "{songObj.get_youtube_link()}"'
                )
                return None

            downloadedFilePathString = await self._perform_audio_download_async(
                convertedFilePath.name, tempFolder, trackAudioStream
            )

            if downloadedFilePathString is None:
                return None

            if dispayProgressTracker:
                dispayProgressTracker.notify_youtube_download_completion()

            downloadedFilePath = Path(downloadedFilePathString)

            ffmpeg_success = await ffmpeg.convert(
                downloaded_file_path=downloadedFilePath,
                converted_file_path=convertedFilePath,
                output_format=self.arguments["format"],
                ffmpeg_path=self.arguments["ffmpeg_path"],
            )

            if dispayProgressTracker:
                dispayProgressTracker.notify_conversion_completion()

            if ffmpeg_success is False:
                # delete the file that wasn't successfully converted
                convertedFilePath.unlink()
            else:
                # if a file was successfully downloaded, tag it
                set_id3_data(convertedFilePath, songObj, self.arguments["format"])

            # Do the necessary cleanup
            if dispayProgressTracker:
                dispayProgressTracker.notify_download_completion()

            if self.downloadTracker:
                self.downloadTracker.notify_download_completion(songObj)

            # delete the unnecessary YouTube download File
            if downloadedFilePath and downloadedFilePath.is_file():
                downloadedFilePath.unlink()

        except Exception as e:
            tb = traceback.format_exc()
            if dispayProgressTracker:
                dispayProgressTracker.notify_error(e, tb)
            else:
                raise e

    async def _perform_audio_download_async(
        self, convertedFileName, tempFolder, trackAudioStream
    ):
        # ! The following function calls blocking code, which would block whole event loop.
        # ! Therefore it has to be called in a separate thread via ThreadPoolExecutor. This
        # ! is not a problem, since GIL is released for the I/O operations, so it shouldn't
        # ! hurt performance.
        return await self.loop.run_in_executor(
            self.thread_executor,
            self._perform_audio_download,
            convertedFileName,
            tempFolder,
            trackAudioStream,
        )

    def _perform_audio_download(self, convertedFileName, tempFolder, trackAudioStream):
        # ! The actual download, if there is any error, it'll be here,
        try:
            # ! pyTube will save the song in .\Temp\$songName.mp4 or .webm,
            # ! it doesn't save as '.mp3'
            downloadedFilePath = trackAudioStream.download(
                output_path=tempFolder, filename=convertedFileName, skip_existing=False
            )
            return downloadedFilePath
        except:  # noqa:E722
            # ! This is equivalent to a failed download, we do nothing, the song remains on
            # ! downloadTrackers download queue and all is well...
            # !
            # ! None is again used as a convenient exit
            tempFiles = Path(tempFolder).glob(f"{convertedFileName}.*")
            for tempFile in tempFiles:
                tempFile.unlink()
            return None



# ===============
# === Imports ===
# ===============

# ! the following are for the search provider to function
import typing

# ! Just for static typing
from typing import List

from rapidfuzz.fuzz import partial_ratio
from ytmusicapi import YTMusic
from bs4 import BeautifulSoup
from requests import get
from unidecode import unidecode


# ================================
# === Note to readers / Coders ===
# ================================

# ! YTM search (the actual POST request), courtesy of Elliot G. (@rocketinventor)
# ! result parsing and song matching system by @Mikhail-Zex
# !
# ! Essentially, Without Elliot, you wouldn't have a YTM search provider at all.


# =======================
# === helper function ===
# =======================


def match_percentage(str1: str, str2: str, score_cutoff: float = 0) -> float:
    """
    `str` `str1` : a random sentence

    `str` `str2` : another random sentence

    `float` `score_cutoff` : minimum score required to consider it a match
                             returns 0 when similarity < score_cutoff

    RETURNS `float`

    A wrapper around `rapidfuzz.fuzz.partial_ratio` to handle UTF-8 encoded
    emojis that usually cause errors
    """

    # ! this will throw an error if either string contains a UTF-8 encoded emoji
    try:
        return partial_ratio(str1, str2, score_cutoff=score_cutoff)

    # ! we build new strings that contain only alphanumerical characters and spaces
    # ! and return the partial_ratio of that
    except:  # noqa:E722
        newStr1 = ""

        for eachLetter in str1:
            if eachLetter.isalnum() or eachLetter.isspace():
                newStr1 += eachLetter

        newStr2 = ""

        for eachLetter in str2:
            if eachLetter.isalnum() or eachLetter.isspace():
                newStr2 += eachLetter

        return partial_ratio(newStr1, newStr2, score_cutoff=score_cutoff)


# ========================================================================
# === Background functions/Variables (Not meant to be called directly) ===
# ========================================================================

# ! YTMusic api client
ytmApiClient = YTMusic()


def _parse_duration(duration: str) -> float:
    """
    Convert string value of time (duration: "25:36:59") to a float value of seconds (92219.0)
    """
    try:
        # {(1, "s"), (60, "m"), (3600, "h")}
        mappedIncrements = zip([1, 60, 3600], reversed(duration.split(":")))
        seconds = 0
        for multiplier, time in mappedIncrements:
            seconds += multiplier * int(time)
        return float(seconds)

    # ! This usually occurs when the wrong string is mistaken for the duration
    except (ValueError, TypeError, AttributeError):
        return 0.0


def _map_result_to_song_data(result: dict) -> dict:
    song_data = {}
    artists = ", ".join(map(lambda a: a["name"], result["artists"]))
    video_id = result["videoId"]
    if video_id is None:
        return {}
    song_data = {
        "name": result["title"],
        "type": result["resultType"],
        "artist": artists,
        "length": _parse_duration(result.get("duration", None)),
        "link": f"https://www.youtube.com/watch?v={video_id}",
        "position": 0,
    }

    album = result.get("album")
    if album:
        song_data["album"] = album["name"]

    return song_data


def _query_and_simplify(searchTerm: str, filter: str) -> List[dict]:
    """
    `str` `searchTerm` : the search term you would type into YTM's search bar
    `str` `filter` : Filter for item types

    RETURNS `list<dict>`

    For structure of dict, see comment at function declaration
    """

    # ! For dict structure, see end of this function (~ln 268, ln 283) and chill, this
    # ! function ain't soo big, there are plenty of comments and blank lines

    # build and POST a query to YTM
    searchResult = ytmApiClient.search(searchTerm, filter=filter)

    return list(map(_map_result_to_song_data, searchResult))


# =======================
# === Search Provider ===
# =======================


def search_and_get_best_match(
    songName: str,
    songArtists: List[str],
    songAlbumName: str,
    songDuration: int,
    isrc: str,
) -> typing.Optional[str]:
    """
    `str` `songName` : name of song

    `list<str>` `songArtists` : list containing name of contributing artists

    `str` `songAlbumName` : name of song's album

    `int` `songDuration` : duration of the song

    `str` `isrc` :  code for identifying sound recordings and music video recordings

    RETURNS `str` : link of the best match
    """

    # if isrc is not None then we try to find song with it
    if isrc is not None:
        isrcResults = _query_and_simplify(isrc, filter="songs")

        if len(isrcResults) == 1:
            isrcResult = isrcResults[0]

            if isrcResult is not None:
                return isrcResult["link"]

    songTitle = create_song_title(songName, songArtists)

    # Query YTM by songs only first, this way if we get correct result on the first try
    # we don't have to make another request to ytmusic api that could result in us
    # getting rate limited sooner
    song_results = _query_and_simplify(songTitle, filter="songs")

    # Order results
    songs = order_ytm_results(
        song_results, songName, songArtists, songAlbumName, songDuration
    )

    # song type results are always more accurate than video type, so if we get score of 80 or above
    # we are almost 100% sure that this is the correct link
    if len(songs) != 0:
        # get the result with highest score
        best_result = max(songs, key=lambda k: songs[k])

        if songs[best_result] >= 80:
            return best_result

    # We didn't find the correct song on the first try so now we get video type results
    # add them to song_results, and get the result with highest score
    video_results = _query_and_simplify(
        create_song_title(songName, songArtists), filter="videos"
    )

    # Order video results
    videos = order_ytm_results(
        video_results, songName, songArtists, songAlbumName, songDuration
    )

    # Merge songs and video results
    results = {**songs, **videos}

    # No matches found
    if len(results) == 0:
        return None

    resultItems = list(results.items())

    # Sort results by highest score
    sortedResults = sorted(resultItems, key=lambda x: x[1], reverse=True)

    # ! In theory, the first 'TUPLE' in sortedResults should have the highest match
    # ! value, we send back only the link
    return sortedResults[0][0]


def order_ytm_results(
    results: List[dict],
    songName: str,
    songArtists: List[str],
    songAlbumName: str,
    songDuration: int,
) -> dict:

    # Assign an overall avg match value to each result
    linksWithMatchValue = {}

    for result in results:
        # ! skip results without videoId, this happens if you are country restricted or
        # ! video is unavailabe
        if result == {}:
            continue

        # ! If there are no common words b/w the spotify and YouTube Music name, the song
        # ! is a wrong match (Like Ruelle - Madness being matched to Ruelle - Monster, it
        # ! happens without this conditional)

        # ! most song results on youtube go by $artist - $songName, so if the spotify name
        # ! has a '-', this function would return True, a common '-' is hardly a 'common
        # ! word', so we get rid of it. Lower-caseing all the inputs is to get rid of the
        # ! troubles that arise from pythons handling of differently cased words, i.e.
        # ! 'Rhino' == 'rhino' is false though the word is same... so we lower-case both
        # ! sentences and replace any hypens(-)
        lowerSongName = songName.lower()
        lowerResultName = result["name"].lower()

        sentenceAWords = lowerSongName.replace("-", " ").split(" ")

        commonWord = False

        # ! check for common word
        for word in sentenceAWords:
            if word != "" and word in lowerResultName:
                commonWord = True

        # ! if there are no common words, skip result
        if not commonWord:
            continue

        # Find artist match
        # ! match  = (no of artist names in result) / (no. of artist names on spotify) * 100
        artistMatchNumber = 0

        # ! we use fuzzy matching because YouTube spellings might be mucked up
        if result["type"] == "song":
            for artist in songArtists:
                if match_percentage(
                    unidecode(artist.lower()), unidecode(result["artist"]).lower(), 85
                ):
                    artistMatchNumber += 1
        else:
            # ! i.e if video
            for artist in songArtists:
                # ! something like match_percentage('rionos', 'aiobahn, rionos Motivation
                # ! (remix)' would return 100, so we're absolutely corrent in matching
                # ! artists to song name.
                if match_percentage(
                    unidecode(artist.lower()), unidecode(result["name"]).lower(), 85
                ):
                    artistMatchNumber += 1

            # we didn't find artist in the video title, so we fallback to
            # detecting song artist in the channel name
            # I am not sure if this won't create false positives
            if artistMatchNumber == 0:
                for artist in songArtists:
                    if match_percentage(
                        unidecode(artist.lower()),
                        unidecode(result["artist"].lower()),
                        85,
                    ):
                        artistMatchNumber += 1

        # ! Skip if there are no artists in common, (else, results like 'Griffith Swank -
        # ! Madness' will be the top match for 'Ruelle - Madness')
        if artistMatchNumber == 0:
            continue

        artistMatch = (artistMatchNumber / len(songArtists)) * 100

        song_title = create_song_title(songName, songArtists)

        # Find name match and drop results below 60%
        # this needs more testing
        if result["type"] == "song":
            nameMatch = round(
                match_percentage(unidecode(result["name"]), unidecode(songName), 60),
                ndigits=3,
            )
        else:
            nameMatch = round(
                match_percentage(unidecode(result["name"]), unidecode(song_title), 60),
                ndigits=3,
            )

        # skip results with name match of 0, these are obviously wrong
        # but can be identified as correct later on due to other factors
        # such as timeMatch or artistMatch
        if nameMatch == 0:
            continue

        # Find album match
        # ! We assign an arbitrary value of 0 for album match in case of video results
        # ! from YouTube Music
        albumMatch = 0.0
        album = None

        if result["type"] == "song":
            album = result.get("album")
            if album:
                albumMatch = match_percentage(album, songAlbumName)

        # Find duration match
        # ! time match = 100 - (delta(duration)**2 / original duration * 100)
        # ! difference in song duration (delta) is usually of the magnitude of a few
        # ! seconds, we need to amplify the delta if it is to have any meaningful impact
        # ! wen we calculate the avg match value
        delta = result["length"] - songDuration
        nonMatchValue = (delta ** 2) / songDuration * 100

        timeMatch = 100 - nonMatchValue

        if result["type"] == "song":
            if album is not None:
                # Don't add albumMatch to avgMatch if songName == result and
                # result album name != songAlbumName
                if (
                    match_percentage(album.lower(), result["name"].lower()) > 95
                    and album.lower() != songAlbumName.lower()
                ):
                    avgMatch = (artistMatch + nameMatch + timeMatch) / 3
                # Add album to avgMatch if songName == result album
                # and result album name == songAlbumName
                else:
                    avgMatch = (artistMatch + albumMatch + nameMatch + timeMatch) / 4
            else:
                avgMatch = (artistMatch + nameMatch + timeMatch) / 3
        # Don't add albumMatch to avgMatch if we don't have information about the album
        # name in the metadata
        else:
            avgMatch = (artistMatch + nameMatch + timeMatch) / 3

        # the results along with the avg Match
        linksWithMatchValue[result["link"]] = avgMatch

    return linksWithMatchValue


def create_song_title(songName: str, songArtists: List[str]) -> str:
    joined_artists = ", ".join(songArtists)
    return f"{joined_artists} - {songName}"


def get_song_lyrics(song_name: str, song_artists: List[str]) -> str:
    """
    `str` `song_name` : name of song

    `list<str>` `song_artists` : list containing name of contributing artists

    RETURNS `str`: Lyrics of the song.

    Gets the metadata of the song.
    """

    headers = {
        "Authorization": "Bearer alXXDbPZtK1m2RrZ8I4k2Hn8Ahsd0Gh_o076HYvcdlBvmc0ULL1H8Z8xRlew5qaG",
    }
    api_search_url = "https://api.genius.com/search"
    search_query = f'{song_name} {", ".join(song_artists)}'

    api_response = get(
        api_search_url, params={"q": search_query}, headers=headers
    ).json()

    song_id = api_response["response"]["hits"][0]["result"]["id"]
    song_api_url = f"https://api.genius.com/songs/{song_id}"

    api_response = get(song_api_url, headers=headers).json()

    song_url = api_response["response"]["song"]["url"]

    genius_page = get(song_url)
    soup = BeautifulSoup(genius_page.text, "html.parser")
    lyrics = soup.select_one("div.lyrics").get_text()

    return lyrics.strip()



from urllib.request import urlopen

from mutagen.easyid3 import EasyID3, ID3
from mutagen.flac import Picture, FLAC
from mutagen.mp4 import MP4, MP4Cover
from mutagen.id3 import APIC as AlbumCover, USLT, COMM as Comment
from mutagen.oggopus import OggOpus
from mutagen.oggvorbis import OggVorbis

import base64

# Apple has specific tags - see mutagen docs -
# http://mutagen.readthedocs.io/en/latest/api/mp4.html
M4A_TAG_PRESET = {
    "album": "\xa9alb",
    "artist": "\xa9ART",
    "date": "\xa9day",
    "title": "\xa9nam",
    "year": "\xa9day",
    "originaldate": "purd",
    "comment": "\xa9cmt",
    "group": "\xa9grp",
    "writer": "\xa9wrt",
    "genre": "\xa9gen",
    "tracknumber": "trkn",
    "albumartist": "aART",
    "discnumber": "disk",
    "cpil": "cpil",
    "albumart": "covr",
    "copyright": "cprt",
    "tempo": "tmpo",
    "lyrics": "\xa9lyr",
    "explicit": "rtng",
}

TAG_PRESET = {}
for key in M4A_TAG_PRESET.keys():
    TAG_PRESET[key] = key


def _set_id3_mp3(converted_file_path, song_obj):
    # embed song details
    # ! we save tags as both ID3 v2.3 and v2.4
    # ! The simple ID3 tags
    audio_file = EasyID3(converted_file_path)

    audio_file = _embed_mp3_metadata(audio_file, song_obj, converted_file_path)

    # ! save as both ID3 v2.3 & v2.4 as v2.3 isn't fully features and
    # ! windows doesn't support v2.4 until later versions of Win10
    audio_file.save(v2_version=3)

    audio_file = _embed_mp3_cover(audio_file, song_obj, converted_file_path)
    audio_file = _embed_mp3_lyrics(audio_file, song_obj)
    # ! setting song links as comment (helpful for devs)
    audio_file.add(Comment(encoding=3, text=song_obj.get_youtube_link()))

    audio_file.save(v2_version=3)


def _set_id3_m4a(converted_file_path, song_obj):
    # embed song details
    # ! we save tags as both ID3 v2.3 and v2.4
    # ! The simple ID3 tags
    audio_file = MP4(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_obj, "m4a", M4A_TAG_PRESET)
    audio_file = _embed_m4a_metadata(audio_file, song_obj)

    audio_file.save()


def _set_id3_flac(converted_file_path, song_obj):
    audio_file = FLAC(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_obj, "flac")
    audio_file = _embed_ogg_metadata(audio_file, song_obj)
    audio_file = _embed_cover(audio_file, song_obj, "flac")

    audio_file.save()


def _set_id3_opus(converted_file_path, song_obj):
    audio_file = OggOpus(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_obj, "opus")
    audio_file = _embed_ogg_metadata(audio_file, song_obj)
    audio_file = _embed_cover(audio_file, song_obj, "opus")

    audio_file.save()


def _set_id3_ogg(converted_file_path, song_obj):
    audio_file = OggVorbis(converted_file_path)

    audio_file = _embed_basic_metadata(audio_file, song_obj, "ogg")
    audio_file = _embed_ogg_metadata(audio_file, song_obj)
    audio_file = _embed_cover(audio_file, song_obj, "ogg")

    audio_file.save()


def _embed_mp3_metadata(audio_file, song_obj, converted_file_path):

    # ! Get rid of all existing ID3 tags (if any exist)
    audio_file.delete()

    # ! song name
    audio_file["title"] = song_obj.get_song_name()
    audio_file["titlesort"] = song_obj.get_song_name()

    # ! track number
    audio_file["tracknumber"] = str(song_obj.get_track_number())

    # ! disc number
    audio_file["discnumber"] = str(song_obj.get_disc_number())

    # ! genres (pretty pointless if you ask me)
    # ! we only apply the first available genre as ID3 v2.3 doesn't support multiple
    # ! genres and ~80% of the world PC's run Windows - an OS with no ID3 v2.4 support
    genres = song_obj.get_genres()
    if len(genres) > 0:
        audio_file["genre"] = genres[0]

    # ! all involved artists
    audio_file["artist"] = song_obj.get_contributing_artists()

    # ! album name
    audio_file["album"] = song_obj.get_album_name()

    # ! album artist (all of 'em)
    audio_file["albumartist"] = song_obj.get_album_artists()

    # ! album release date (to what ever precision available)
    audio_file["date"] = song_obj.get_album_release()
    audio_file["originaldate"] = song_obj.get_album_release()

    return audio_file


def _embed_mp3_cover(audio_file, song_obj, converted_file_path):
    # ! setting the album art
    audio_file = ID3(converted_file_path)
    rawAlbumArt = urlopen(song_obj.get_album_cover_url()).read()
    audio_file["APIC"] = AlbumCover(
        encoding=3, mime="image/jpeg", type=3, desc="Cover", data=rawAlbumArt
    )

    return audio_file


def _embed_mp3_lyrics(audio_file, song_obj):
    # ! setting the lyrics
    lyrics = song_obj.get_lyrics()
    USLTOutput = USLT(encoding=3, lang=u"eng", desc=u"desc", text=lyrics)
    audio_file["USLT::'eng'"] = USLTOutput

    return audio_file


def _embed_m4a_metadata(audio_file, song_obj):
    # set year
    years = song_obj.get_album_release().split("-")
    if len(years) >= 1:
        audio_file[M4A_TAG_PRESET["year"]] = years[0]

    # set youtube link as comment
    youtube_link = song_obj.get_youtube_link()
    if youtube_link:
        audio_file[M4A_TAG_PRESET["comment"]] = youtube_link

    # set lyrics
    lyrics = song_obj.get_lyrics()
    if lyrics:
        audio_file[M4A_TAG_PRESET["lyrics"]] = lyrics

    # Explicit values: Dirty: 4, Clean: 2, None: 0
    audio_file[M4A_TAG_PRESET["explicit"]] = (0,)
    try:
        audio_file[M4A_TAG_PRESET["albumart"]] = [
            MP4Cover(
                urlopen(song_obj.get_album_cover_url()).read(),
                imageformat=MP4Cover.FORMAT_JPEG,
            )
        ]
    except IndexError:
        pass

    return audio_file


def _embed_basic_metadata(audio_file, song_obj, encoding, preset=TAG_PRESET):

    # set main artist
    artists = song_obj.get_contributing_artists()
    if len(artists):
        audio_file[preset["artist"]] = artists[0]
        audio_file[preset["albumartist"]] = artists[0]

    # set song title
    song_title = song_obj.get_song_name()
    if song_title:
        audio_file[preset["title"]] = song_title

    # set album name
    album_name = song_obj.get_album_name()
    if album_name:
        audio_file[preset["album"]] = album_name

    # set release data
    release_data = song_obj.get_album_release()
    if release_data:
        audio_file[preset["date"]] = release_data
        audio_file[preset["originaldate"]] = release_data

    # set genre
    genres = song_obj.get_genres()
    if len(genres) >= 1:
        audio_file[preset["genre"]] = genres[0]

    # set disc number
    disc_number = song_obj.get_disc_number()
    if disc_number:
        if encoding in ["flac", "ogg", "opus"]:
            audio_file[preset["discnumber"]] = str(disc_number)
        else:
            audio_file[preset["discnumber"]] = [(disc_number, 0)]

    # set track number
    track_number = song_obj.get_track_number()
    if track_number:
        if encoding in ["flac", "ogg", "opus"]:
            audio_file[preset["tracknumber"]] = str(track_number)
        else:
            audio_file[preset["tracknumber"]] = [(track_number, 0)]

    return audio_file


def _embed_ogg_metadata(audio_file, song_obj):
    # set year
    years = song_obj.get_album_release().split("-")
    if len(years) >= 1:
        audio_file["year"] = years[0]

    # set youtube link as comment
    youtube_link = song_obj.get_youtube_link()
    if youtube_link:
        audio_file["comment"] = youtube_link

    # set lyrics
    lyrics = song_obj.get_lyrics()
    if lyrics:
        audio_file["lyrics"] = lyrics

    return audio_file


def _embed_cover(audio_file, song_obj, encoding):
    image = Picture()
    image.type = 3
    image.desc = "Cover"
    image.mime = "image/jpeg"
    image.data = urlopen(song_obj.get_album_cover_url()).read()

    if encoding == "flac":
        audio_file.add_picture(image)
    elif encoding == "ogg" or encoding == "opus":
        # From the Mutagen docs (https://mutagen.readthedocs.io/en/latest/user/vcomment.html)
        image_data = image.write()
        encoded_data = base64.b64encode(image_data)
        vcomment_value = encoded_data.decode("ascii")
        audio_file["metadata_block_picture"] = [vcomment_value]

    return audio_file


SET_ID3_FUNCTIONS = {
    "mp3": _set_id3_mp3,
    "flac": _set_id3_flac,
    "opus": _set_id3_opus,
    "ogg": _set_id3_ogg,
    "m4a": _set_id3_m4a,
}


def set_id3_data(converted_file_path, song_obj, outputFormat):
    function = SET_ID3_FUNCTIONS.get(outputFormat)
    if function:
        function(converted_file_path, song_obj)



# ===============
# === Imports ===
# ===============

# ! These are not used, they're here for static type checking using mypy
from spotdl.search.songObj import SongObj
import spotdl.search.songGatherer as songGatherer
import typing
from pathlib import Path


# ===============================
# === Download tracking class ===
# ===============================


class DownloadTracker:
    def __init__(self):
        self.songObjList = []
        self.saveFile: typing.Optional[Path] = None

    def load_tracking_file(self, trackingFilePath: str) -> None:
        """
        `str` `trackingFilePath` : path to a .spotdlTrackingFile

        RETURNS `~`

        reads songsObj's from disk and prepares to track their download
        """

        # Attempt to read .spotdlTrackingFile, raise exception if file can't be read
        trackingFile = Path(trackingFilePath)
        if not trackingFile.is_file():
            raise FileNotFoundError(f"no such tracking file found: {trackingFilePath}")

        with trackingFile.open("rb") as file_handle:
            songDataDumps = eval(file_handle.read().decode())

        # Save path to .spotdlTrackingFile
        self.saveFile = trackingFile

        # convert song data dumps to songObj's
        # ! see, songGatherer.get_data_dump and songGatherer.from_dump for more details
        for dump in songDataDumps:
            self.songObjList.append(songGatherer.from_dump(dump))

    def load_song_list(self, songObjList: typing.List[SongObj]) -> None:
        """
        `list<songOjb>` `songObjList` : songObj's being downloaded

        RETURNS `~`

        prepares to track download of provided songObj's
        """

        self.songObjList = songObjList

        self.backup_to_disk()

    def get_song_list(self) -> typing.List[SongObj]:
        """
        RETURNS `list<songObj>

        get songObj's representing songs yet to be downloaded
        """
        return self.songObjList

    def backup_to_disk(self):
        """
        RETURNS `~`

        backs up details of songObj's yet to be downloaded to a .spotdlTrackingFile
        """
        # remove tracking file if no songs left in queue
        # ! we use 'return None' as a convenient exit point
        if len(self.songObjList) == 0:
            if self.saveFile and self.saveFile.is_file():
                self.saveFile.unlink()
            return None

        # prepare datadumps of all songObj's yet to be downloaded
        songDataDumps = []

        for song in self.songObjList:
            songDataDumps.append(song.get_data_dump())

        # ! the default naming of a tracking file is $nameOfFirstSOng.spotdlTrackingFile,
        # ! it needs a little fixing because of disallowed characters in file naming
        if not self.saveFile:
            songName = self.songObjList[0].get_song_name()

            for disallowedChar in ["/", "?", "\\", "*", "|", "<", ">"]:
                if disallowedChar in songName:
                    songName = songName.replace(disallowedChar, "")

            songName = songName.replace('"', "'").replace(":", " - ")

            self.saveFile = Path(songName + ".spotdlTrackingFile")

        # backup to file
        # ! we use 'wb' instead of 'w' to accommodate your fav K-pop/J-pop/Viking music
        with open(self.saveFile, "wb") as file_handle:
            file_handle.write(str(songDataDumps).encode())

    def notify_download_completion(self, songObj: SongObj) -> None:
        """
        `songObj` `songObj` : songObj representing song that has been downloaded

        RETURNS `~`

        removes given songObj from download queue and updates .spotdlTrackingFile
        """

        if songObj in self.songObjList:
            self.songObjList.remove(songObj)

        self.backup_to_disk()

    def clear(self):
        self.songObjList = []
        self.saveFile = None



from spotdl.search.songObj import SongObj
from spotdl.search.spotifyClient import SpotifyClient
import spotdl.search.audioProvider as audioProvider
import spotdl.search.metadataProvider as metadataProvider

from typing import List, Dict
from pathlib import Path

# =======================
# === Master Gatherer ===
# =======================


def from_query(request: str, output_format: str = None):
    songObjList = []
    if "open.spotify.com" in request and "track" in request:
        print("Fetching Song...")
        songObjList = [songobj_from_spotify_url(request, output_format)]

    elif "open.spotify.com" in request and "album" in request:
        print("Fetching Album...")
        songObjList = get_album_tracks(request, output_format)

    elif "open.spotify.com" in request and "playlist" in request:
        print("Fetching Playlist...")
        songObjList = get_playlist_tracks(request, output_format)

    elif "open.spotify.com" in request and "artist" in request:
        print("Fetching artist...")
        songObjList = get_artist_tracks(request, output_format)

    elif request == "saved":
        print("Fetching Saved Songs...")
        songObjList = get_saved_tracks(output_format)

    else:
        print('Searching Spotify for song named "%s"...' % request)
        try:
            songObjList = from_search_term(request, output_format)
        except Exception as e:
            print(e)

    # filter out NONE songObj items (already downloaded)
    songObjList = [songObj for songObj in songObjList if songObj]

    return songObjList


# All other functions in this file call this one


def songobj_from_spotify_url(spotifyURL: str, output_format: str = None):
    # check if URL is a playlist, user, artist or album, if yes raise an Exception,
    # else procede

    # Get the Song Metadata
    print(f"Gathering Spotify Metadata for: {spotifyURL}")
    rawTrackMeta, rawArtistMeta, rawAlbumMeta = metadataProvider.from_url(spotifyURL)

    songName = rawTrackMeta["name"]
    albumName = rawTrackMeta["album"]["name"]
    isrc = rawTrackMeta["external_ids"].get("isrc")
    contributingArtists = []
    for artist in rawTrackMeta["artists"]:
        contributingArtists.append(artist["name"])
    duration = round(rawTrackMeta["duration_ms"] / 1000, ndigits=3)

    convertedFileName = SongObj.create_file_name(
        songName, [artist["name"] for artist in rawTrackMeta["artists"]]
    )

    if len(convertedFileName) > 250:
        convertedFileName = SongObj.create_file_name(
            songName, [rawTrackMeta["artists"][0]['name']]
        )

    if output_format is None:
        output_format = "mp3"

    convertedFilePath = Path(".", f"{convertedFileName}.{output_format}")
    displayName = ", ".join(contributingArtists) + " - " + songName

    # if a song is already downloaded skip it
    if convertedFilePath.is_file():
        print(f'Skipping "{convertedFileName}" as it\'s already downloaded')
        return None

    # Get the song's downloadable audio link
    print(f'Searching YouTube for "{displayName}"', end="\r")
    youtubeLink = audioProvider.search_and_get_best_match(
        songName, contributingArtists, albumName, duration, isrc
    )
    if youtubeLink is None:
        # raise Exception("Could not match any of the results on YouTube")
        print("Could not match any of the results on YouTube. Skipping")
        return None
    else:
        print(" " * (len(displayName) + 25), end="\r")
        print(f'Found YouTube URL for "{displayName}" : {youtubeLink}')

    # (try to) Get lyrics from Genius
    try:
        lyrics = metadataProvider.get_song_lyrics(songName, contributingArtists)
    except (AttributeError, IndexError):
        lyrics = ""

    return SongObj(rawTrackMeta, rawAlbumMeta, rawArtistMeta, youtubeLink, lyrics)


# =======================
# === Other Gatherers ===
# =======================


def from_dump(dataDump: dict):
    rawTrackMeta = dataDump["rawTrackMeta"]
    rawAlbumMeta = dataDump["rawAlbumMeta"]
    rawArtistMeta = dataDump["rawAlbumMeta"]
    youtubeLink = dataDump["youtubeLink"]
    lyrics = dataDump["lyrics"]

    return SongObj(rawTrackMeta, rawAlbumMeta, rawArtistMeta, youtubeLink, lyrics)


def from_search_term(query: str, output_format: str = None) -> List[SongObj]:
    """
    `str` `query` : what you'd type into Spotify's search box
    Queries Spotify for a song and returns the best match
    """

    # get a spotify client
    spotifyClient = SpotifyClient()

    # get possible matches from spotify
    result = spotifyClient.search(query, type="track")

    # return first result link or if no matches are found, raise and Exception
    if len(result["tracks"]["items"]) == 0:
        raise Exception("No song matches found on Spotify")
    else:
        songUrl = "http://open.spotify.com/track/" + result["tracks"]["items"][0]["id"]
        song = songobj_from_spotify_url(songUrl, output_format)
        return [song] if song is not None else []


def get_album_tracks(albumUrl: str, output_format: str = None) -> List[SongObj]:
    """
    `str` `albumUrl` : Spotify Url of the album whose tracks are to be
    retrieved
    returns a `list<songObj>` containing Url's of each track in the given album
    """

    spotifyClient = SpotifyClient()

    albumResponse = spotifyClient.album_tracks(albumUrl)
    albumTracks = albumResponse["items"]
    tracks = []

    # Get all tracks from album
    while albumResponse["next"]:
        albumResponse = spotifyClient.next(albumResponse)
        albumTracks.extend(albumResponse["items"])

    # Create song objects from track ids
    for track in albumTracks:
        song = songobj_from_spotify_url(
            "https://open.spotify.com/track/" + track["id"], output_format
        )

        if song is not None and song.get_youtube_link() is not None:
            tracks.append(song)

    return tracks


def get_artist_tracks(artistUrl: str, output_format: str = None) -> List[SongObj]:
    """
    `str` `albumUrl` : Spotify Url of the artist whose tracks are to be
    retrieved
    returns a `list<songObj>` containing Url's of each track in the artist profile
    """

    spotifyClient = SpotifyClient()

    artistTracks = []

    artistResponse = spotifyClient.artist_albums(artistUrl, album_type="album,single")
    albumsList = artistResponse["items"]
    albumsObject: Dict[str, str] = {}
    # Fetch all artist albums
    while artistResponse["next"]:
        artistResponse = spotifyClient.next(artistResponse)
        albumsList.extend(artistResponse["items"])

    # Remove duplicate albums
    for album in albumsList:
        # return an iterable containing the string's alphanumeric characters
        alphaNumericFilter = filter(str.isalnum, album["name"].lower())

        # join all characters into one string
        albumName = "".join(alphaNumericFilter)

        if albumsObject.get(albumName) is None:
            albumsObject[albumName] = album["uri"]

    tracksList = []
    tracksObject: Dict[str, str] = {}
    # Fetch all tracks from all albums
    for album_uri in albumsObject.values():
        albumResponse = spotifyClient.album_tracks(album_uri)
        albumTracks = albumResponse["items"]

        while albumResponse["next"]:
            spotifyClient.next(albumResponse)
            albumTracks.extend(albumResponse["items"])

        tracksList.extend(albumTracks)

    # Filter tracks to remove duplicates and songs without our artist
    for track in tracksList:
        # return an iterable containing the string's alphanumeric characters
        alphaNumericFilter = filter(str.isalnum, track["name"].lower())

        # join all characters into one string
        trackName = "".join(alphaNumericFilter)

        if tracksObject.get(trackName) is None:
            for artist in track["artists"]:
                # get artist id from url
                # https://api.spotify.com/v1/artists/1fZAAHNWdSM5gqbi9o5iEA/albums
                #
                # split string
                #  ['https:', '', 'api.spotify.com', 'v1', 'artists',
                #  '1fZAAHNWdSM5gqbi9o5iEA', 'albums']
                #
                # get second element from the end
                # '1fZAAHNWdSM5gqbi9o5iEA'
                artistId = artistResponse["href"].split("/")[-2]

                # ignore tracks that are not from our artist by checking
                # the id
                if artist["id"] == artistId:
                    tracksObject[trackName] = track["uri"]

    # Create song objects from track ids
    for trackUri in tracksObject.values():
        song = songobj_from_spotify_url(
            f"https://open.spotify.com/track/{trackUri}", output_format
        )

        if song is not None and song.get_youtube_link() is not None:
            artistTracks.append(song)

    return artistTracks


def get_playlist_tracks(playlistUrl: str, output_format: str = None) -> List[SongObj]:
    """
    `str` `playlistUrl` : Spotify Url of the album whose tracks are to be
    retrieved
    returns a `list<songObj>` containing Url's of each track in the given playlist
    """

    spotifyClient = SpotifyClient()
    playlistResponse = spotifyClient.playlist_tracks(playlistUrl)
    playlistTracks = playlistResponse["items"]
    tracks = []

    # while loop to mimic do-while
    while playlistResponse["next"]:
        playlistResponse = spotifyClient.next(playlistResponse)
        playlistTracks.extend(
            [
                track
                for track in playlistResponse["items"]
                # check if track has id
                if track.get("track", {}).get("id") is not None
            ]
        )

    for track in playlistTracks:
        song = songobj_from_spotify_url(
            "https://open.spotify.com/track/" + track["track"]["id"],
            output_format,
        )

        if song is not None and song.get_youtube_link() is not None:
            tracks.append(song)

    return tracks


def get_saved_tracks(output_format: str = None) -> List[SongObj]:
    """
    returns a `list<songObj>` containing Url's of each track in the user's saved tracks
    """

    spotifyClient = SpotifyClient()

    savedTracksResponse = spotifyClient.current_user_saved_tracks()
    savedTracks = savedTracksResponse["items"]
    tracks = []

    while savedTracksResponse["next"]:
        savedTracksResponse = spotifyClient.next(savedTracksResponse)
        savedTracks.extend(
            [
                track
                for track in savedTracksResponse["items"]
                # check if track has id
                if track.get("track", {}).get("id") is not None
            ]
        )

    for track in savedTracks:
        song = songobj_from_spotify_url(
            "https://open.spotify.com/track/" + track["track"]["id"],
            output_format,
        )

        if song is not None and song.get_youtube_link() is not None:
            tracks.append(song)

    return tracks



from requests import get
from bs4 import BeautifulSoup
from spotdl.search.spotifyClient import SpotifyClient
from typing import List


def from_url(spotifyURL: str):
    if not ("open.spotify.com" in spotifyURL and "track" in spotifyURL):
        raise Exception("passed URL is not that of a track: %s" % spotifyURL)

    # query spotify for song, artist, album details
    spotifyClient = SpotifyClient()

    rawTrackMeta = spotifyClient.track(spotifyURL)

    primaryArtistId = rawTrackMeta["artists"][0]["id"]
    rawArtistMeta = spotifyClient.artist(primaryArtistId)

    albumId = rawTrackMeta["album"]["id"]
    rawAlbumMeta = spotifyClient.album(albumId)

    return rawTrackMeta, rawArtistMeta, rawAlbumMeta


def get_song_lyrics(song_name: str, song_artists: List[str]) -> str:
    """
    `str` `song_name` : name of song

    `list<str>` `song_artists` : list containing name of contributing artists

    RETURNS `str`: Lyrics of the song.

    Gets the metadata of the song.
    """

    headers = {
        "Authorization": "Bearer alXXDbPZtK1m2RrZ8I4k2Hn8Ahsd0Gh_o076HYvcdlBvmc0ULL1H8Z8xRlew5qaG",
    }
    api_search_url = "https://api.genius.com/search"
    search_query = f'{song_name} {", ".join(song_artists)}'

    api_response = get(
        api_search_url, params={"q": search_query}, headers=headers
    ).json()

    song_id = api_response["response"]["hits"][0]["result"]["id"]
    song_api_url = f"https://api.genius.com/songs/{song_id}"

    api_response = get(song_api_url, headers=headers).json()

    song_url = api_response["response"]["song"]["url"]

    genius_page = get(song_url)
    soup = BeautifulSoup(genius_page.text, "html.parser")
    lyrics = soup.select_one("div.lyrics").get_text()

    return lyrics.strip()
